import os
import sys
from loguru import logger


def get_logger():
    logger.remove()
    log_format = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"
    logger.add(sys.stdout, format=log_format, level="DEBUG")
    return logger



from multiprocessing.managers import BaseManager
from collections import deque
from typing import List, Any


class Dispatcher:
    def __init__(self, max_size: int = 120):
        self.queue = deque(maxlen=max_size)
        self.current_time = 0
        self.current_frame = 0
        self.collect_results = {}

    def get_size(self) -> int:
        return len(self.queue)

    def add_frame(self, frame):
        self.queue.append(frame)
        self.current_frame += 1

    def get_frames(self, count: int, step: int) -> List[Any]:
        if len(self.queue) < count * step:
            return []

        return list(self.queue)[-count * step :][::-1][::step][::-1]

    def collect_result(self, inference_result):
        if inference_result is not None:
            time = str(inference_result[0])
            name = inference_result[1]
            data = inference_result[2]
            if self.collect_results.get(name) is None:
                self.collect_results[name] = {}
            self.collect_results[name][time] = data

    def clear_collect_results(self):
        self.collect_results.clear()

    def clear(self):
        self.queue.clear()
        self.collect_results.clear()
        self.current_time = 0
        self.current_frame = 0

    def increase_current_time(self):
        self.current_time += 1

    def get_current_time(self) -> int:
        return self.current_time

    def set_current_time(self, time):
        self.current_time = time

    def get_current_frame(self) -> int:
        return self.current_frame

    def set_current_frame(self, frame):
        self.current_frame = frame

    @classmethod
    def create(cls, max_size=120, offline=False):
        if offline:
            return cls(max_size)
        else:
            return DispatcherManager(cls).create(max_size)


class DispatcherManager:
    def __init__(self, obj=None):
        self._manager = None
        self._dispatcher = None
        self._obj = Dispatcher if obj is None else obj

    def create(self, max_size: int = 120):
        if self._manager is None:
            self._initialize_manager(max_size)
        return self._dispatcher

    def _initialize_manager(self, max_size):
        BaseManager.register("Dispatcher", self._obj)
        self._manager = BaseManager()
        self._manager.start()
        self._dispatcher = self._manager.Dispatcher(max_size)



import cv2

from ..log import logger


class OpenCVProducer:
    def __init__(self, width: int, height: int, cvt_code=None):
        self.width = width
        self.height = height
        self.cvt_code = cvt_code

    def read(self, path, fps=None, position=0):
        """
        Reads frames from a video file/stream_url/v4l2 device.
        Optionally skips frames to meet the specified fps.

        Args:
            path (str): The path to the video file/stream_url/v4l2 device.
            fps (int, optional): Target frames per second. If None, no frame skipping is done.
            position (int, optional): The position in seconds from where to start reading the video.

        Yields:
            numpy.ndarray: frame
        """
        cap = cv2.VideoCapture(path)
        if not cap.isOpened():
            raise ValueError(f"Failed to open {path}")

        original_fps = cap.get(cv2.CAP_PROP_FPS)

        # Skip to the requested second
        if position > 0:
            cap.set(cv2.CAP_PROP_POS_MSEC, position * 1000)  # position in milliseconds

        frame_interval = 1.0
        if fps is not None and original_fps > fps:
            frame_interval = original_fps / fps

        frame_index = int(original_fps * position)
        next_frame_to_process = frame_index
        while True:
            ret, frame = cap.read()
            if not ret:
                break

            if frame_index >= next_frame_to_process:
                try:
                    height, width, _ = frame.shape
                    if width != self.width or height != self.height:
                        frame = cv2.resize(frame, (self.width, self.height))

                    if self.cvt_code is not None:
                        frame = cv2.cvtColor(frame, self.cvt_code)

                    yield frame
                    next_frame_to_process += frame_interval
                except Exception as e:
                    logger.error(f"Error processing frame: {e}")
                    raise e

            frame_index += 1

        cap.release()

    def get_info(self, path):
        """
        Extracts video properties.

        Args:
            path (str): The path to the video file/stream_url/v4l2 device.

        Returns:
            dict: Video properties including width, height, fps, and frame count.
        """
        cap = cv2.VideoCapture(path)
        if not cap.isOpened():
            raise ValueError(f"Failed to open {path}")

        width = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        fps = cap.get(cv2.CAP_PROP_FPS)
        frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        cap.release()

        return {
            "width": width,
            "height": height,
            "fps": fps,
            "frame_count": frame_count,
        }



def trans_position2time(position: int) -> str:
    return f"{position // 3600:02d}:{position // 60 % 60:02d}:{position % 60:02d}"


def trans_time2position(time: str) -> int:
    h, m, s = time.split(":")
    return int(h) * 3600 + int(m) * 60 + int(s)



import os
import streamlit as st
from .inference import Inference
from .player import Player
from .producer import OpenCVProducer, PyAVProducer
from .util import trans_position2time
from .log import logger


class StreamInferApp:
    def __init__(self, inference: Inference, save_output_path: str = "./outputs"):
        if not isinstance(inference, Inference):
            logger.error("inference must be an instance of Inference")
            return
        self.inference = inference
        self.save_output_path = save_output_path
        self.setup_ui()
        self.setup_output_ui()
        if self.need_save:
            if not os.path.exists(save_output_path):
                os.makedirs(save_output_path)
        self.annotate_frame_func = self.default_annotate_frame
        self.output_func = self.default_output

    def setup_output_ui(self):
        if not self.show_output:
            return
        self.output_widgets = {}
        algos = self.inference.list_algos()
        for idx, tab in enumerate(st.tabs(algos)):
            self.output_widgets[algos[idx]] = tab.empty()

    def setup_ui(self):
        st.set_page_config(
            page_title="StreamInfer",
            page_icon="📸",
            layout="wide",
            initial_sidebar_state="expanded",
        )

        with st.sidebar:
            st.header("参数控制")
            self.video_path = st.text_input(
                "流URL或视频路径",
                placeholder="网络流地址/绝对路径",
                disabled=self.get_state("input_disabled", bool),
            )
            res_width, res_height = st.columns(2)
            self.width = res_width.number_input(
                "推理宽度", 320, 1920, 1920, disabled=self.get_state("input_disabled", bool)
            )
            self.height = res_height.number_input(
                "推理高度", 180, 1080, 1080, disabled=self.get_state("input_disabled", bool)
            )
            self.frame_rate = st.slider(
                "帧率", 1, 60, 30, disabled=self.get_state("input_disabled", bool)
            )
            self.color_channel = st.selectbox(
                "颜色通道", ("BGR", "RGB"), disabled=self.get_state("input_disabled", bool)
            )
            self.start_position = st.number_input(
                "起始播放位置(秒)", 0, 1000, 0, disabled=self.get_state("input_disabled", bool)
            )
            st.divider()

            self.show_frame = st.toggle(
                "展示视频帧", value=True, disabled=self.get_state("control_disabled", bool)
            )
            self.show_output = st.toggle(
                "展示推理数据", disabled=self.get_state("control_disabled", bool)
            )
            self.need_save = st.toggle(
                "保存推理结果", value=True, disabled=self.get_state("control_disabled", bool)
            )

            st.button(
                "开始推理",
                on_click=self.start_infer,
                use_container_width=True,
                disabled=self.get_state("control_disabled", bool),
            )
            if self.get_state("mode") == "infering":
                st.button(
                    "停止/返回推理",
                    on_click=self.stop_infer,
                    use_container_width=True,
                    type="primary",
                )

        st.header("推理结果")
        if self.need_save:
            st.caption("结果将保存到 ./outputs 文件夹")
        if self.show_frame:
            self.image_frame = st.image([])
        self.video_progress = st.progress(0, text="视频待加载")

    def get_state(self, key, key_type=str):
        if key not in st.session_state:
            st.session_state[key] = key_type()
        return st.session_state[key]

    def set_state(self, key, value):
        st.session_state[key] = value

    def start_infer(self):
        if not self.video_path:
            st.error("请输入视频路径")
            return
        self.set_state("control_disabled", True)
        self.set_state("input_disabled", True)
        self.set_state("mode", "infering")

    def stop_infer(self):
        self.set_state("control_disabled", False)
        self.set_state("input_disabled", False)
        self.set_state("mode", "standby")

    def fmt_output(self, result):
        output = ""
        for key, data in result.items():
            output += f"第 {key} 秒\n"
            output += "-------------------------------\n"
            output += f"{data}\n"
            output += "=======================================\n\n\n"
        return output

    def save_output(self):
        for algo_name in self.inference.list_algos():
            result = self.inference.dispatcher.get_result(algo_name)
            if result is None:
                result = ""
            path = os.path.join(self.save_output_path, f"{algo_name}.log")
            with open(path, "w+", encoding="utf-8") as f:
                f.write(self.fmt_output(result))

    def run_inference(self, use_opencv: bool = True, clear: bool = False):
        if use_opencv:
            producer = OpenCVProducer(self.width, self.height)
        else:
            producer = PyAVProducer(self.width, self.height)
        player = Player(self.inference.dispatcher, producer, path=self.video_path)

        video_info = player.info
        is_realtime = video_info["frame_count"] <= 0
        if not is_realtime:
            total_sec = int(video_info["frame_count"] / video_info["fps"])
            total_sec_display = trans_position2time(total_sec)
            self.video_progress.progress(0, text=f"00:00:00 / {total_sec_display}")
            temp_current_time = 0
        else:
            self.video_progress.progress(1, text="实时视频流")

        for frame, current_frame in player.play(
            self.frame_rate, position=self.start_position
        ):
            now_current_time = self.inference.dispatcher.get_current_time()
            if not is_realtime and now_current_time != temp_current_time:
                temp_current_time = now_current_time
                self.video_progress.progress(
                    temp_current_time / total_sec,
                    text=f"{trans_position2time(temp_current_time)} / {total_sec_display}",
                )

            current_algo_names = self.inference.auto_run_specific(
                player.play_fps, current_frame
            )
            if self.show_frame:
                for algo in self.inference.list_algos():
                    _, data = self.inference.dispatcher.get_last_result(algo)
                    if data is None:
                        continue
                    frame = self.annotate_frame_func(algo, data, frame)
            if len(current_algo_names) == 0:
                yield None, -1, None, frame
            else:
                for idx, current_algo_name in enumerate(current_algo_names):
                    position, data = self.inference.dispatcher.get_last_result(
                        current_algo_name
                    )
                    if idx != len(current_algo_names) - 1:
                        display_frame = None
                    else:
                        display_frame = frame
                    yield current_algo_name, position, data, display_frame

        if self.need_save:
            self.save_output()
        self.stop_infer()
        if clear:
            self.inference.dispatcher.clear()
            st.rerun()
        else:
            st.stop()

    def default_annotate_frame(self, name, data, frame):
        return frame

    def set_annotate_frame(self, func):
        def custom_annotate_frame_wrapper(name, data, frame):
            return func(self, name, data, frame)

        self.annotate_frame_func = custom_annotate_frame_wrapper

    def default_output(self, name, position, data):
        self.output_widgets[name].text(f"{position}: {data}")

    def set_output(self, func):
        def custom_output_wrapper(name, position, data):
            return func(self, name, position, data)

        self.output_func = custom_output_wrapper

    def start(self, use_opencv: bool = True, clear: bool = False):
        if self.get_state("mode") != "infering":
            st.stop()

        for current_algo_name, position, data, frame in self.run_inference(
            use_opencv=use_opencv, clear=clear
        ):
            if self.show_frame and frame is not None:
                self.image_frame.image(frame, channels=self.color_channel)
            if self.show_output and current_algo_name:
                self.output_func(current_algo_name, position, data)



import av
import cv2

from ..log import logger


class PyAVProducer:
    def __init__(self, width: int, height: int, format=None):
        self.width = width
        self.height = height
        self.format = "bgr24" if format is None else format

    def read(self, path, fps=None, position=0):
        """
        Reads frames from a video file/stream_url/v4l2 device.
        Optionally skips frames to meet the specified fps.

        Args:
            path (str): The path to the video file/stream_url/v4l2 device.
            fps (int, optional): Target frames per second. If None, no frame skipping is done.
            position (int, optional): The position in seconds from where to start reading the video.

        Yields:
            numpy.ndarray: frame
        """
        try:
            container = av.open(path)
            video_stream = next(s for s in container.streams if s.type == "video")
            original_fps = video_stream.base_rate

            # Seek to the specified position
            if position > 0:
                logger.warning(
                    "Using PyAVProducer and specifying position is not recommended because there is not yet a good solution to the problem of startup delays but it still works"
                )
                start_frame = int(position * original_fps)

            frame_interval = 1.0
            if fps is not None and original_fps > fps:
                frame_interval = original_fps / fps

            frame_index = 0
            next_frame_to_process = start_frame if position > 0 else frame_index
            for frame in container.decode(video=0):
                if frame_index >= next_frame_to_process:
                    try:
                        frame = frame.to_ndarray(format=self.format)
                        height, width, _ = frame.shape
                        if width != self.width or height != self.height:
                            frame = cv2.resize(frame, (self.width, self.height))

                        yield frame
                        next_frame_to_process += frame_interval
                    except Exception as e:
                        logger.error(f"Error processing frame: {e}")
                        raise e

                frame_index += 1

        except av.AVError as e:
            logger.error(f"Failed to open {path}: {e}")
            raise ValueError(f"Failed to open {path}: {e}")

        container.close()

    def get_info(self, path):
        """
        Extracts video properties.

        Args:
            path (str): The path to the video file/stream_url/v4l2 device.

        Returns:
            dict: Video properties including width, height, fps, and frame count.
        """
        try:
            container = av.open(path)
            video_stream = next(s for s in container.streams if s.type == "video")

            width = video_stream.width
            height = video_stream.height
            fps = video_stream.base_rate  # or video_stream.average_rate

            if hasattr(video_stream, "frames"):
                frame_count = video_stream.frames
            else:
                frame_count = 0

            return {
                "width": width,
                "height": height,
                "fps": fps,
                "frame_count": frame_count,
            }

        except av.AVError as e:
            raise ValueError(f"Failed to open {path}: {e}")



import imageio
import cv2

from ..log import logger


class ImageIOProducer:
    def __init__(self, width: int, height: int, cvt_code=None):
        self.width = width
        self.height = height
        self.cvt_code = cvt_code

    def read(self, path, fps=None):
        """
        Reads frames from a video file.

        Args:
            path (str): The path to the video file.
                e.g., "/path/to/video.mp4"

        Yields:
            numpy.ndarray: frame
        """
        try:
            reader = imageio.get_reader(path)
        except Exception as e:
            logger.error(f"Error reading video: {e}")
            raise e

        original_fps = reader.get_meta_data()["fps"]

        # Calculate the frame skip interval as a float if fps is set and original_fps is higher
        frame_interval = 1.0
        if fps is not None and original_fps > fps:
            frame_interval = original_fps / fps

        frame_index = 0
        next_frame_to_process = 0
        for frame in reader:
            try:
                height, width, _ = frame.shape
                if width != self.width or height != self.height:
                    frame = cv2.resize(frame, (self.width, self.height))

                if self.cvt_code is not None:
                    frame = cv2.cvtColor(frame, self.cvt_code)

                # cv2.imwrite("frame.jpg", frame)
                yield frame
                next_frame_to_process += (
                    frame_interval  # Update the next frame index to process
                )
            except Exception as e:
                logger.error(f"Error processing frame: {e}")
                raise e

            frame_index += 1

    def get_info(self, path):
        reader = imageio.get_reader(path)
        fps = reader.get_meta_data()["fps"]
        frame_count = reader.count_frames()

        frame = reader.get_next_data()
        height, width, _ = frame.shape

        reader.close()

        return {
            "width": width,
            "height": height,
            "fps": fps,
            "frame_count": frame_count,
        }



import av
import os
import cv2
from enum import Enum
from .producer import OpenCVProducer, PyAVProducer
from .player import Player


class RecordMode(Enum):
    NONE = 0
    OPENCV = 1
    PYAV = 2


class Recorder:
    def __init__(self, player: Player, recording_path: str):
        self.record_mode = RecordMode.NONE
        self.player = player
        self.width = player.producer.width
        self.height = player.producer.height
        self.fps = player.play_fps
        self.recording_path = self._ensure_mp4_extension(recording_path)
        self._initialize_writer()

    def _ensure_mp4_extension(self, path):
        if not path.endswith(".mp4"):
            path += ".mp4"
        os.makedirs(os.path.dirname(path), exist_ok=True)
        return path

    def _initialize_writer(self):
        if isinstance(self.player.producer, OpenCVProducer):
            self._init_opencv_writer()
        elif isinstance(self.player.producer, PyAVProducer):
            self._init_pyav_writer()

    def _init_opencv_writer(self):
        self.record_mode = RecordMode.OPENCV
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        self.writer = cv2.VideoWriter(
            self.recording_path, fourcc, self.fps, (self.width, self.height)
        )

    def _init_pyav_writer(self):
        self.record_mode = RecordMode.PYAV
        try:
            container = av.open(self.recording_path, mode="w")
        except av.AVError as e:
            # Handle AVError if necessary
            raise e
        stream = container.add_stream("h264", rate=self.fps)
        stream.width, stream.height = (self.width, self.height)
        stream.pix_fmt = "yuv420p"
        self.container = container
        self.stream = stream

    def add_frame(self, frame):
        try:
            if self.record_mode == RecordMode.OPENCV:
                self.writer.write(frame)
            elif self.record_mode == RecordMode.PYAV:
                self._add_frame_pyav(frame)
        except Exception as e:
            # Handle exceptions related to frame addition
            raise e

    def _add_frame_pyav(self, frame):
        frame = av.VideoFrame.from_ndarray(frame, format=self.player.producer.format)
        frame.pict_type = "NONE"
        for packet in self.stream.encode(frame):
            self.container.mux(packet)

    def close(self):
        try:
            if self.record_mode == RecordMode.OPENCV:
                self.writer.release()
            elif self.record_mode == RecordMode.PYAV:
                self.container.close()
        except Exception as e:
            # Handle exceptions during closing
            raise e



import time
import multiprocessing as mp
from multiprocessing.managers import BaseProxy

from .util import trans_position2time
from .log import logger


class Player:
    def __init__(self, dispatcher, producer, path):
        self.dispatcher = dispatcher
        self.producer = producer
        self.path = path
        try:
            self.info = self.producer.get_info(self.path)
        except Exception as e:
            raise ValueError(f"Error getting info: {e}")
        self.fps = self.info["fps"]
        self.play_fps = self.fps
        self.frame_count = self.info["frame_count"]
        self.process = None
        self.is_end = mp.Value("b", False)

    def play(self, fps=None, position=0):
        fps = self.fps if fps is None else fps
        self.play_fps = fps
        interval_count = 0

        if position > 0:
            self.dispatcher.set_current_time(position)
            self.dispatcher.set_current_frame(fps * position)

        for idx, frame in enumerate(self.producer.read(self.path, fps, position)):
            self.dispatcher.add_frame(frame)
            interval_count += 1
            if interval_count >= fps:
                interval_count = 0
                self.dispatcher.increase_current_time()
                logger.debug(f"current time: {self.get_play_time()}")
            yield frame, self.dispatcher.get_current_frame()

    def play_async(self, fps=None):
        """
        Starts the appropriate streaming process based on the frame count.
        """
        if not isinstance(self.dispatcher, BaseProxy):
            logger.error(
                f"Dispatcher is not an proxy: {type(self.dispatcher)}, use create(offline=False) to create"
            )
            raise ValueError(
                f"Dispatcher is not an proxy: {type(self.dispatcher)}, use create(offline=False) to create"
            )

        if fps is None or fps >= self.fps:
            fps = self.fps
            if fps > 30:
                logger.warning(
                    f"FPS {fps} is too high, if your player is playing more slowly than the actual time, set a lower play fps"
                )
        self.play_fps = fps

        if self.frame_count <= 0:
            target = self.normal_stream
        else:
            target = self.video_stream

        self.process = mp.Process(target=target)
        self.process.start()
        return self.process

    def stop(self):
        if self.process:
            self.is_end.value = True
        self.process.terminate()

    def is_active(self) -> bool:
        """
        Checks if the streaming process is still running.
        """
        return (
            self.process.is_alive() and not self.is_end.value if self.process else False
        )

    def get_play_time(self) -> str:
        return trans_position2time(self.dispatcher.get_current_time())

    def video_stream(self):
        """
        Handles streaming for video files. Frames are processed at a rate determined by the video's FPS.
        """
        base_interval = 1 / self.play_fps
        start_time = time.time()
        interval_count = 0

        for idx, frame in enumerate(self.producer.read(self.path, self.play_fps)):
            target_time = start_time + (idx * base_interval)
            time.sleep(max(0, target_time - time.time()))
            self.dispatcher.add_frame(frame)
            interval_count += 1
            if interval_count >= self.play_fps:
                interval_count = 0
                self.dispatcher.increase_current_time()
                logger.debug(f"current time: {self.get_play_time()}")

        self.is_end.value = True

    def normal_stream(self):
        """
        Handles streaming for non-video files. Frames are processed at regular intervals.
        """
        for frame in self.producer.read(self.path, self.play_fps):
            if self.dispatcher.get_current_frame() % self.play_fps == 0:
                self.dispatcher.increase_current_time()
                logger.debug(f"current time: {self.get_play_time()}")
            self.dispatcher.add_frame(frame)

        self.is_end.value = True



from .base import Dispatcher
from ..log import logger


class DevelopDispatcher(Dispatcher):
    """
    For develop
    """

    def get_result(self, name, clear=False):
        if self.collect_results.get(name):
            data = self.collect_results[name].copy()
            if clear:
                del self.collect_results[name]
            return data
        return None

    def get_last_result(self, name, clear=False):
        algo_results = self.get_result(name, clear)
        if algo_results is not None and len(algo_results.keys()) > 0:
            time_point = int(max([int(k) for k in algo_results.keys()]))
            return time_point, algo_results[str(time_point)]
        return -1, None



import threading as th
import cv2

from .player import Player
from .recorder import Recorder
from .timer import Timer
from .log import logger


class Inference:
    def __init__(self, dispatcher):
        self.dispatcher = dispatcher
        self.inferences_info = []
        self.timers = {}
        self.is_stop = False
        self.process_func = self.default_process

    def load_algo(self, algo_instance, frame_count, frame_step, interval):
        self.inferences_info.append((algo_instance, frame_count, frame_step, interval))
        self.timers[algo_instance.name] = Timer(interval)
        algo_instance.init()

    def list_algos(self):
        result = []
        for algo_instance, _, _, _ in self.inferences_info:
            result.append(algo_instance.name)
        return result

    def run(self):
        for inference_info in self.inferences_info:
            algo_instance, _, _, _ = inference_info
            timer = self.timers[algo_instance.name]
            if timer.is_time():
                self._infer(inference_info)

    def run_loop(self):
        while not self.is_stop:
            self.run()

    def run_async(self):
        thread = th.Thread(target=self.run_loop)
        thread.start()
        return thread

    def stop(self):
        self.is_stop = True

    def auto_run_specific(self, fps, current_frame) -> str:
        current_algo_names = []
        for algo_instance, _, _, frequency in self.inferences_info:
            if current_frame % int(frequency * fps) == 0:
                self.run_specific(algo_instance.name)
                current_algo_names.append(algo_instance.name)
        return current_algo_names

    def run_specific(self, algo_name):
        for inference_info in self.inferences_info:
            algo_instance, _, _, _ = inference_info
            if algo_instance.name == algo_name:
                self._infer(inference_info)

    def _infer(self, inference_info):
        algo_instance, frame_count, frame_step, _ = inference_info
        frames = self.dispatcher.get_frames(frame_count, frame_step)
        if not frames:
            return -1
        result = algo_instance.run(frames)
        self.dispatcher.collect_result(
            (self.dispatcher.get_current_time(), algo_instance.name, result)
        )
        return result

    def default_process(self, *args, **kwargs):
        pass

    def set_custom_process(self, func):
        def custom_process_wrapper(*args, **kwargs):
            return func(self, *args, **kwargs)

        self.process_func = custom_process_wrapper

    def start(
        self,
        player: Player,
        fps: int = 30,
        position: int = 0,
        offline: bool = False,
        recording_path: str = None,
    ):
        if offline:
            recorder = Recorder(player, recording_path) if recording_path else None
            for frame, current_frame in player.play(fps, position):
                current_algo_names = self.auto_run_specific(fps, current_frame)
                processed_frame = self.process_func(
                    frame=frame, current_algo_names=current_algo_names
                )
                frame = processed_frame if processed_frame is not None else frame
                if recorder:
                    recorder.add_frame(frame)
            if recorder:
                recorder.close()
        else:
            player.play_async(fps)
            self.run_async()
            while player.is_active():
                self.process_func()
            self.stop()
            player.stop()
        self.dispatcher.clear()



class BaseAlgo:
    def __init__(self, name=None) -> None:
        if not name:
            name = self.__class__.__name__
        self.name = name

    def init(self):
        """
        Initialize the algo model or other resources.
        """
        raise NotImplementedError

    def run(self):
        """
        Run inference and return the inference result.
        """
        raise NotImplementedError



import time


class Timer:
    def __init__(self, interval):
        self.interval = interval
        self.last_call = None

    def wait(self):
        current_time = time.time()

        if self.last_call is not None:
            elapsed = current_time - self.last_call
            if elapsed > self.interval:
                pass
            else:
                time.sleep(self.interval - elapsed)

        self.last_call = time.time()

    def is_time(self) -> bool:
        current_time = time.time()
        if self.last_call is None or current_time - self.last_call >= self.interval:
            self.last_call = current_time
            return True
        return False
